//! Provides efficient concatenation of strings and vectors
//!
//! The goal of these macros are to reduce the amount of allocations that are required
//! when concatenating string buffers and vectors; with a macro that makes it simple to
//! achieve in practice.
//!
//! # Implementation Notes
//!
//! - The caller must provide a buffer for appending the slices to
//! - The buffer is resized to accomodate the total length of all slices given

/// Appends any number of string slices onto a string buffer
///
/// ```rust
/// use concat_in_place::strcat;
///
/// let domain = "domain.com";
/// let endpoint = "inventory/parts";
/// let id = "10512";
///
/// let mut url = String::new();
/// let slice = strcat!(&mut url, domain "/" endpoint "/" id);
/// assert_eq!(slice, "domain.com/inventory/parts/10512");
/// ```
///
/// # Implementation Notes
///
/// Technically works with any string type that has the following methods:
///
/// - `capacity`
/// - `len`
/// - `push_str`
#[macro_export]
macro_rules! strcat {
    ($input:expr, $($element:expr)*) => {{
        let out = $input;
        let mut required = 0;

        $(
            required += $element.len();
        )*

        let free = out.capacity() - out.len();
        if (free < required) {
            out.reserve(required - free);
        }

        $(
            out.push_str($element);
        )*

        &*out
    }}
}

/// Appends any number of slices onto a vector
///
/// ```rust
/// use concat_in_place::veccat;
///
/// let domain = b"domain.com";
/// let endpoint = b"inventory/parts";
/// let id = b"10512";
///
/// let mut url = Vec::new();
/// let slice = veccat!(&mut url, domain b"/" endpoint b"/" id);
/// assert_eq!(slice, b"domain.com/inventory/parts/10512");
/// ```
///
/// # Implementation Notes
///
/// Technically works with any type that has the following methods:
///
/// - `capacity`
/// - `len`
/// - `reserve`
/// - `extend_from_slice`
#[macro_export]
macro_rules! veccat {
    ($input:expr, $($element:expr)*) => {{
        let out = $input;
        let mut required = 0;

        $(
            required += $element.len();
        )*

        let free = out.capacity() - out.len();
        if (free < required) {
            out.reserve(required - free);
        }

        $(
            out.extend_from_slice($element);
        )*

        &*out
    }}
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn bytes() {
        let mut out = Vec::new();

        let expected = b"domain.com/endpoint/id";
        let actual: &[u8] = veccat!(&mut out, b"domain.com" b"/" b"endpoint" b"/" b"id");
        assert_eq!(actual, expected, "{}", String::from_utf8_lossy(actual));
    }

    #[test]
    fn strings() {
        let mut out = String::new();

        let expected = "domain.com/endpoint/id";
        let actual = strcat!(&mut out, "domain.com" "/" "endpoint" "/" "id");
        assert_eq!(actual, expected);
    }
}
